#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

"""
The purpose of this will to collect all the ltms.
build a supergraph of all the machines
evaluate if time range needs to be higher than machine name to allow bitmasks across them.
                                                alcf
                                                machine names
                                    theta                                   polaris
                                            time ranges
                        r0            r1              r2                r0              r1
                               racks, slots, rows and other branches could be here.
                    n0 n1..        n0 n1 n2..       n1 n2..        n1 n2 n3..       ng1 ng2 ng3..

"""
import importlib
import os
import sys
from datetime import datetime
from pprint import pformat
from typing import Type, Optional, List

import numpy as np
import pandas as pd
from pandas import Timestamp

from Octeres.SL.abstract_location_to_mask import (
    get_pk_hash,
    AbstractLocationtoMaskError,
    Abstract_Location_to_Mask,
)
from Octeres.SL.location_to_mask_graph import (
    XLocation_to_Mask_None,
    XLocation_to_Mask_DoesNotExist,
)
from Octeres.bitmask import EnhancedBitmask, EnhancedBitmaskLike
from Octeres.logical import create_mask
from Octeres.forthwith import EPOCH_START_NOTZ, EPOCH_END_NOTZ
from Octeres.util import superprint


def prepare_ltm_module_lst(sl_modules_path):
    """prepare and get all the possible ltm modules"""
    possible_ltms = []
    default_sl_modules_path, current_file = os.path.split(__file__)
    possible_ltms.extend(list(filter(lambda p: p.startswith("ltm_"), os.listdir(default_sl_modules_path))))
    possible_ltms.extend(list(filter(lambda p: p.startswith("ltm_"), os.listdir(sl_modules_path))))
    ltm_module_lst = []
    for ltm_modules in possible_ltms:
        module_name = ltm_modules.replace(".py", "")
        machine_name = module_name.replace("ltm_", "")
        ltm_module_lst.append([module_name, machine_name])
    if os.path.exists(sl_modules_path) and sl_modules_path not in sys.path:
        sys.path.append(sl_modules_path)
    if os.path.exists(default_sl_modules_path) and default_sl_modules_path not in sys.path:
        sys.path.append(default_sl_modules_path)
    return ltm_module_lst


def init_ltm(ser: pd.Series, bitmask_class: type[EnhancedBitmaskLike], debug=False):
    ltm_class: type[Abstract_Location_to_Mask] = ser["ltm_class"]
    ltm_obj = ltm_class(bitmask_class=bitmask_class, debug=debug)
    if ltm_class not in (XLocation_to_Mask_None, XLocation_to_Mask_DoesNotExist):
        ts = pd.to_datetime(ser["ts"]).to_pydatetime()  # .isoformat()
        te = pd.to_datetime(ser["te"]).to_pydatetime()  # .isoformat()
        alias = ser["alias"]
        notes = ser["notes"]
        test_only = ser["test_only"]
        pk = ser.name
        ltm_obj.set_init(pk, ts, te)
        ltm_obj.notes = notes
        ltm_obj.alias = alias
    return ltm_obj


class Super_Location_to_Mask:
    """The Dragon Slayer"""

    _df: pd.DataFrame
    _instance = None
    sl_modules_path = None
    bitmask_class: type[EnhancedBitmaskLike] = None

    @staticmethod
    def get_index_columns_and_dtypes():
        df_index = ["pk"]
        column_lst = [
            "ts",
            "te",
            "ltm_class",
            "name",
            "pk",
            "alias",
            "notes",
            "test_only",
            "bitmask_class",
        ]
        dtypes = {
            #'ts': Timestamp, 'te': Timestamp,
            #'ltm_class': Abstract_Location_to_Mask,
            "name": "string",
            # 'pk': np.int64,
            "alias": "string",
            "notes": "string",
            "test_only": np.int8,
            #'bitmask_class': EnhancedBitmaskLike,
        }
        defaults = {
            "name": "",
            "alias": "",
            "notes": "",
            "test_only": 0,
        }
        return df_index, column_lst, dtypes, defaults

    @classmethod
    def get_DataFrame(cls, ltm_module_lst):
        df_index, column_lst, dtypes, defaults = cls.get_index_columns_and_dtypes()
        for dct in ltm_module_lst:
            for key, value in defaults.items():
                if key not in dct:
                    dct[key] = value
        df = pd.DataFrame(ltm_module_lst, columns=column_lst)
        df = df.astype(dtypes)
        df = df.set_index(df_index)
        return df

    def __new__(cls, sl_modules_path=""):
        # this returns a singleton.
        if cls._instance is None:
            if not sl_modules_path:
                cls.sl_modules_path = os.environ.get("OCTERES_SL_MODULES_PATH", None)
            else:
                cls.sl_modules_path = sl_modules_path

            if not cls.sl_modules_path:
                default_sl_modules_path, current_file = os.path.split(__file__)
                cls.sl_modules_path = default_sl_modules_path
                # raise Exception(f"Pass a sl_modules_path to the creation of Super_Location_to_Mask or set OCTERES_SL_MODULES_PATH.")

            cls._instance = super(Super_Location_to_Mask, cls).__new__(cls)
            cls._df = cls.get_DataFrame([])
        return cls._instance

    def set_bitmask_class(self, bitmask_class):
        self.bitmask_class = bitmask_class

    def set_sl_modules_path(self, sl_modules_path):
        self.sl_modules_path = sl_modules_path

    def get_ltm_df_logical(self, logical_filt, bitmask_class=None, include_test=False):
        """return a filtered data frame of ltms using a logical filter.
        See Octeres.logical
        example:
            logical_filt = {
                "&": [
                    {'name': 'name', 'op': '==', 'val': 'test'},
                    {'name': 'alias', 'op': '==', 'val': 'test_14_n'}
                     ]
                }
        """
        bitmask_class = self.bitmask_class if bitmask_class is None else bitmask_class
        # df = self.get_ltm_df({}, bitmask_class=bitmask_class, include_test=include_test)
        df = self._df
        df = df.reset_index()
        mask = create_mask(df, logical_filt)
        df = df[mask]
        df = df.loc[df["bitmask_class"] == bitmask_class]
        if not include_test:
            df = df.loc[df["test_only"] != 1]
        return df

    def get_ltm_df(self, filt, bitmask_class=None, include_test=False):
        """return a filtered data frame of ltms.  This is the old way of filtering."""
        bitmask_class = self.bitmask_class if bitmask_class is None else bitmask_class
        df = self._df
        filt_keys = filt.keys()
        if filt == {}:
            pass
        elif "name" in filt_keys:
            df = df.loc[df["name"] == filt["name"]]
            if "pit" in filt_keys:
                pd_point_in_time = pd.to_datetime(filt["pit"])
                df = df.loc[(df["ts"] <= pd_point_in_time) & (pd_point_in_time < df["te"])]
            elif "ts" in filt_keys and "te" in filt_keys:
                ts = pd.to_datetime(filt["ts"])
                te = pd.to_datetime(filt["te"])
                # df = ltm_df.loc[(ltm_df['ts'] < te) & (ltm_df['te'] > ts)]
                if ts == te:
                    # odd case where we are selecting a single exact record
                    df = df.loc[(df["ts"] == te) & (df["te"] == ts)]
                else:
                    df = df.loc[(df["ts"] < te) & (df["te"] > ts)]
        else:
            raise AbstractLocationtoMaskError(f"unsupported filter {filt}.")
        if bitmask_class is None:
            df = df.loc[df["bitmask_class"].isna()]
        else:
            df = df.loc[df["bitmask_class"] == bitmask_class]
        if not include_test:
            df = df.loc[df["test_only"] != 1]
        return df

    def slay(self, filt: Optional[dict] = None, bitmask_class=None, include_test=True):
        """
        Slay the Dragon!
        Iterate through the modules and join all the ltms together.
        Validate some of the ltm module variables and classes.
        """
        # print(f"slay {filt=}")
        filt = filt or {}
        df = self.get_ltm_df(filt, include_test=include_test, bitmask_class=bitmask_class)
        # If get_Super_Location_to_Mask is called more than once, the second time may not include the machine you request later
        # because you don't have a filt.
        if df.empty or filt == {}:
            missing_all = []
            ltm_module_lst = prepare_ltm_module_lst(self.sl_modules_path)
            for module_name, machine_name in ltm_module_lst:
                # print(module_name, machine_name)
                if filt:
                    if type(filt) == dict:
                        if machine_name != filt["name"]:
                            continue
                    else:
                        raise Exception(f"Invalid filt type({type(filt)}).")
                find_dct = {"MACHINE_NAME": 0, "ltm_lst": 0}
                ltm_module = importlib.import_module(module_name, package="")
                ltm_module_contents = [(getattr(ltm_module, x), x) for x in dir(ltm_module)]
                ltm_classes = []
                for obj, name in ltm_module_contents:
                    if not (name.startswith("--") or name.endswith("__")):
                        if name in find_dct:
                            find_dct[name] += 1
                        if name == "ltm_lst":
                            for ltm_module_dct in obj:
                                if ltm_module_dct["ltm_class"] in [
                                    XLocation_to_Mask_DoesNotExist,
                                    XLocation_to_Mask_None,
                                ]:
                                    pk = get_pk_hash(
                                        ltm_module_dct["name"],
                                        ltm_module_dct["ts"],
                                        ltm_module_dct["te"],
                                    )
                                    ltm_module_dct["pk"] = pk
                                ltm_module_dct["bitmask_class"] = bitmask_class
                                self.insert_ltm_class_range(ltm_module_dct)
                                ltm_class = ltm_module_dct["ltm_class"]
                                if ltm_class == XLocation_to_Mask_None:
                                    continue
                                ltm_classes.append(ltm_module_dct)
                missing = []
                for key, value in find_dct.items():
                    if value < 1:
                        missing.append(key)
                if missing:
                    missing_all.append(f"{module_name} missing: {','.join(missing)}")
            if missing_all:
                for missing in missing_all:
                    # TODO: evaluate raising
                    print(f"Missing:{missing}")
        #
        # df['alias'] = df['alias'].fillna("")
        # df['notes'] = df['notes'].fillna("")
        # df['test_only'] = df['test_only'].fillna(0)

    def insert_ltm_class_range(self, ltm_module_dct):
        """add a ltm to the dataframe."""
        df: pd.DataFrame = self._df
        change = False
        bitmask_class = ltm_module_dct["bitmask_class"]
        try:
            pk = ltm_module_dct["pk"]
        except KeyError:
            new_pk = get_pk_hash(ltm_module_dct["name"], ltm_module_dct["ts"], ltm_module_dct["te"])
            raise AbstractLocationtoMaskError(f"Missing pk, suggest: {new_pk} for {pformat(ltm_module_dct)}")
        df_exists: pd.DataFrame = df.loc[((df.index == pk) & (df.bitmask_class == bitmask_class))]
        if df_exists.empty:
            ltm_df = self.get_DataFrame([ltm_module_dct])
            if not df.empty and not ltm_df.empty:
                df = pd.concat([df, ltm_df], ignore_index=False)
            elif not df.empty and ltm_df.empty:
                df = df
            else:
                df = ltm_df
            change = True
        else:
            # we found one, lets make sure they are equal
            ltm_df = self.get_DataFrame([ltm_module_dct])
            if not df_exists.equals(ltm_df):
                new_pk = get_pk_hash(ltm_module_dct["name"], ltm_module_dct["ts"], ltm_module_dct["te"])
                print(f"duplicated pk, please evaluate pk: {pk}, suggest: {new_pk} for {pformat(ltm_module_dct)}")
        if change:
            df["alias"] = df["alias"].fillna("")
            df["notes"] = df["notes"].fillna("")
            df["test_only"] = df["test_only"].fillna(0)
            self._df = df

    def ltm_object_factory_filter(self, logical_filt, bitmask_class=None, include_test=False):
        df = self.get_ltm_df_logical(logical_filt, bitmask_class=bitmask_class, include_test=include_test)
        ltm_obj_lst = []
        for pk, ser in df.iterrows():
            if ser["ltm_class"] in (
                XLocation_to_Mask_None,
                XLocation_to_Mask_DoesNotExist,
            ):
                continue
            ltm_obj = init_ltm(ser, self.bitmask_class)
            ltm_obj_lst.append(ltm_obj)
        return ltm_obj_lst

    def ltm_object_factory_filter_single(self, logical_filt, bitmask_class=None, include_test=False):
        ltm_obj_lst = self.ltm_object_factory_filter(logical_filt, bitmask_class=bitmask_class, include_test=include_test)
        if len(ltm_obj_lst) == 1:
            ltm_obj = ltm_obj_lst.pop()
        elif len(ltm_obj_lst) > 1:
            raise AbstractLocationtoMaskError(f"The logical filter return ed too many ltms:{len(ltm_obj_lst)=}")
        else:
            ltm_class = XLocation_to_Mask_DoesNotExist
            ltm_obj = ltm_class()
        return ltm_obj

    def ltm_object_factory(
        self,
        machine_name,
        range_start: datetime,
        range_end: Optional[datetime] = None,
        bitmask_class=None,
        include_test=False,
        debug=False,
    ) -> Abstract_Location_to_Mask:
        """given a machine name, range_start and range_end, return a single ltm.
        Since it's using ltm_class_factory_by_range, it will use only range_start to
        search for the ltm."""
        bitmask_class = self.bitmask_class if bitmask_class is None else bitmask_class
        if range_start and range_end:
            pd_range_start = pd.to_datetime(range_start)
            pd_range_end = pd.to_datetime(range_end)
            df = self.get_ltm_df(
                {"name": machine_name, "ts": pd_range_start, "te": pd_range_end},
                bitmask_class=bitmask_class,
                include_test=include_test,
            )
        else:
            if range_start:
                point_in_time = range_start
            elif range_end:
                # if we are inserting a ltm with an initial time and range end, range_start will be Null.
                point_in_time = range_end
            else:
                raise AbstractLocationtoMaskError(f"Invalid LTM at with range {range_start} -> {range_end}")
            pd_point_in_time = pd.to_datetime(point_in_time)
            # fixme: cannot get a point_in_time that is 0 seconds long.
            #  While invalid, still possible in the data.
            df = self.get_ltm_df(
                {"name": machine_name, "pit": pd_point_in_time},
                bitmask_class=bitmask_class,
                include_test=include_test,
            )
        if len(df) == 1:
            ser = df.iloc[0]
            ltm_class = ser["ltm_class"]
            if ltm_class not in [
                XLocation_to_Mask_DoesNotExist,
                XLocation_to_Mask_None,
            ]:
                ltm_obj = init_ltm(ser, bitmask_class, debug=debug)
            else:
                raise AbstractLocationtoMaskError(f"{machine_name=} {range_start=} {range_end=} not found")
        elif len(df) > 1:
            raise AbstractLocationtoMaskError(f"{machine_name=} {range_start=} {range_end=} returned too many ltms:{len(df)=}")
        else:
            raise AbstractLocationtoMaskError(f"{machine_name=} {range_start=} {range_end=} not found")
        return ltm_obj

    def ltm_objects_factory(
        self,
        machine_name,
        range_start,
        range_end,
        bitmask_class=None,
        include_test=False,
    ) -> List[Abstract_Location_to_Mask]:
        """given a machine name, range_start and range_end, return all the ltms in that range."""
        if range_start is None:  # search for the last/up to range_end
            df = self.get_ltm_df(
                {"name": machine_name, "pit": range_end},
                bitmask_class=bitmask_class,
                include_test=include_test,
            )
        else:
            df = self.get_ltm_df(
                {"name": machine_name, "ts": range_start, "te": range_end},
                bitmask_class=bitmask_class,
                include_test=include_test,
            )
        # FIXME: test if sort_values and df.iterrows maintain order. link to 29627134
        # FIXME: it may be better to add an order in the dataframe.
        df = df.sort_values(by=["ts"])
        ltm_obj_lst = []
        for pk, ser in df.iterrows():
            if ser["ltm_class"] in (
                XLocation_to_Mask_None,
                XLocation_to_Mask_DoesNotExist,
            ):
                continue
            ltm_obj = init_ltm(ser, self.bitmask_class)
            ltm_obj_lst.append(ltm_obj)
        ltm_obj_lst.sort(key=lambda ltm_obj: ltm_obj.range_start)  # FIXME: remove when tested.  link to 29627134
        return ltm_obj_lst

    def ltm_objects_factory_all(self, machine_name, bitmask_class=None, include_test=False):
        """return all the ltms for the bitmask_class"""
        return self.ltm_objects_factory(
            machine_name,
            EPOCH_START_NOTZ,
            EPOCH_END_NOTZ,
            bitmask_class=bitmask_class,
            include_test=include_test,
        )

    def ltm_get_time_range(self, machine_name, bitmask_class=bitmask_class, include_test=False):
        """get a ltms time range."""
        df = self.get_ltm_df(bitmask_class=bitmask_class, include_test=include_test)
        valid_df = df.loc[
            (machine_name == df["name"])
            & (XLocation_to_Mask_None != df["ltm_class"])
            & (XLocation_to_Mask_DoesNotExist != df["ltm_class"])
        ]
        range_start = valid_df["ts"].min()
        range_end = valid_df["te"].max()
        return range_start, range_end


def get_Super_Location_to_Mask(
    filt: Optional[dict] = None,
    bitmask_class: type[EnhancedBitmaskLike] = EnhancedBitmask,
):
    """return the SLTM singleton."""
    filt = filt or {}
    sltm = Super_Location_to_Mask()
    sltm.set_bitmask_class(bitmask_class)
    sltm.slay(filt=filt, bitmask_class=bitmask_class)
    return sltm


def get_BitTree(logical_filt: dict, bitmask_class: type[EnhancedBitmaskLike] = EnhancedBitmask):
    """
    get a bitmask tree that acts like a location to mask.
    """
    sltm = get_Super_Location_to_Mask(bitmask_class=bitmask_class)
    sltm.set_bitmask_class(bitmask_class)
    df = sltm._df
    mask = create_mask(df, logical_filt)
    df = df[mask]
    return df
