# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

from itertools import permutations
import networkx as nx
from pandas import to_datetime as pddt
from Octeres.SL.location_to_mask_graph import XLocation_to_Mask_None
from Octeres.SL.abstract_location_to_mask import (
    graph_from_name_to_bitmask_dct,
    ABC_Location_to_Mask_Graph,
)
from Octeres.SL.obj_type import Obj_Type
from Octeres.bitmask import EnhancedBitmask as eb, EnhancedBitmaskLike, EnhancedBitmask
from Octeres.bitmask.bitmask_graph_search import (
    apply_bitmask_to_nxgraph,
    is_bitmask_indexable,
)

_, MACHINE_NAME = __name__.rsplit("ltm_", 1)

"""
binary DIGraph
DIGraph with parents with n n-1 and n+1
Non-uniform DIGraph

When naming, I put the MSB to the left.
"""


class Location_to_Mask_Simple(ABC_Location_to_Mask_Graph):
    machine_name = MACHINE_NAME
    obj_types = [Obj_Type.ROOT.value, Obj_Type.BLOCK.value]
    bit_total = 8
    graph_only = True
    names = [
        (
            ("switch", (0, 1, 1), "switch"),
            ("node", (0, 8, 1), "node{node:0>2}"),
        )
    ]

    def get_parents(self, machine_name: str, node_name: str) -> list:
        if self.graph:
            parents = super().get_parents(machine_name, node_name)
        elif node_name == machine_name:
            parents = []
        elif node_name == "switch":
            parents = [machine_name]
        else:
            parents = ["switch"]
        return parents

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        if self.graph:
            obj_type = super().get_obj_type(machine_name, node_name)
        elif machine_name == node_name:
            obj_type = Obj_Type.ROOT.value
        elif node_name == "switch":
            obj_type = Obj_Type.SWITCH.value
        else:
            obj_type = Obj_Type.NODE.value
        return obj_type


class Location_to_Mask_Binary(ABC_Location_to_Mask_Graph):
    machine_name = MACHINE_NAME
    obj_types = [Obj_Type.ROOT.value, Obj_Type.BLOCK.value]
    bit_total = 16
    graph_only = True
    names = [
        (
            ("l0", (0, 2, 1), "b-{l0:0>2}"),
            ("l1", (0, 2, 1), "b-{l0:0>2}-{l1:0>2}"),
            ("l2", (0, 2, 1), "b-{l0:0>2}-{l1:0>2}-{l2:0>2}"),
            ("l3", (0, 2, 1), "b-{l0:0>2}-{l1:0>2}-{l2:0>2}-{l3:0>2}"),
        )
    ]

    def get_parents(self, machine_name: str, node_name: str) -> list:
        if self.graph:
            parents = super().get_parents(machine_name, node_name)
        elif node_name == machine_name:
            parents = []
        else:
            prefix = "b-"
            name_parts = node_name.replace(prefix, "").split("-")
            if len(name_parts) == 1:
                parents = [machine_name]
            else:
                parent = "b-" + "-".join(name_parts[:-1])
                parents = [parent]
        return parents

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        if self.graph:
            obj_type = super().get_obj_type(machine_name, node_name)
        elif machine_name == node_name:
            obj_type = Obj_Type.ROOT.value
        else:
            obj_type = Obj_Type.BLOCK.value
        return obj_type


class Location_to_Mask_BinaryArray(ABC_Location_to_Mask_Graph):
    machine_name = MACHINE_NAME
    obj_types = [Obj_Type.ROOT.value, Obj_Type.BLOCK.value]
    bit_total = 16
    graph_only = True
    names_extra = {"bit_total": bit_total}
    name_to_bitmask_dct = {
        machine_name: eb.array(
            [
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
            ]
        ),
        "b0000": eb.array([1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0001": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1]),
        "b0002": eb.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0003": eb.array([0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0004": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0]),
        "b0005": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1]),
        "b0006": eb.array([1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0007": eb.array([0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0008": eb.array([0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0009": eb.array([0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0010": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0]),
        "b0011": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0]),
        "b0012": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0]),
        "b0013": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1]),
        "b0014": eb.array([1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0015": eb.array([0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0016": eb.array([0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0017": eb.array([0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0018": eb.array([0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0019": eb.array([0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0020": eb.array([0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0021": eb.array([0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0]),
        "b0022": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0]),
        "b0023": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0]),
        "b0024": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0]),
        "b0025": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0]),
        "b0026": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0]),
        "b0027": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0]),
        "b0028": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0]),
        "b0029": eb.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1]),
    }

    def get_parents(self, machine_name: str, node_name: str) -> list:
        if self.graph:
            parents = super().get_parents(machine_name, node_name)
        elif node_name == machine_name:
            parents = []
        else:
            parents = []
            bitmask = self.name_to_bitmask_dct[node_name]
            bit_count = bitmask.bit_count
            for onode_name, obitmask in self.name_to_bitmask_dct.items():
                if node_name == onode_name:
                    continue
                if (bit_count * 2) != obitmask.bit_count:
                    continue
                if eb.logical_is_subset(obitmask, bitmask):
                    parents.append(onode_name)
            if not parents:
                raise Exception("parent not found.")
        return parents

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        if self.graph:
            obj_type = super().get_obj_type(machine_name, node_name)
        elif machine_name == node_name:
            obj_type = Obj_Type.ROOT.value
        else:
            obj_type = Obj_Type.BLOCK.value
        return obj_type

    def generate_bittree(
        self,
        validate=False,
        bitmask_class: type[EnhancedBitmaskLike] = EnhancedBitmask,
        debug=False,
    ):
        if bitmask_class != EnhancedBitmask:
            name_to_bitmask_dct = {}
            for key, value in self.name_to_bitmask_dct.items():
                idx_lst = eb.get_idx_lst_from_bitmask(value)
                name_to_bitmask_dct[key] = bitmask_class.get_bitmask_from_idx_lst(value.length, idx_lst)
        else:
            name_to_bitmask_dct = self.name_to_bitmask_dct
        graph = graph_from_name_to_bitmask_dct(self, name_to_bitmask_dct)
        return graph


def func(name_dct, generate_parent=False):
    idx = name_dct["idx"]  # index at level
    lvl = name_dct["lvl"]
    l0 = name_dct.get("l0", None)
    l1 = name_dct.get("l1", None)
    l2 = name_dct.get("l2", None)
    l3 = name_dct.get("l3", None)
    bit_total = name_dct["bit_total"]
    level_tuple = (l0, l1, l2, l3)
    level_count = len(level_tuple)
    rlvl = level_count - lvl - 1
    offset = sum([int(bit_total / (2**i)) for i in range(0, rlvl)])
    name = offset + idx
    name = f"b{name:0>4}"
    # print(name_dct)
    print(
        f"{name: >4}, {str(idx)=: >4}, lvl:{str(lvl): >4}, rlvl:{str(rlvl): >4}, {str(l0): >4}, {str(l1): >4}, {str(l2): >4}, {str(l3): >4}"
    )
    return name


class Location_to_Mask_BinaryGraph(ABC_Location_to_Mask_Graph):
    machine_name = MACHINE_NAME
    obj_types = [Obj_Type.ROOT.value, Obj_Type.BLOCK.value]
    bit_total = 8
    graph_only = True
    graph_nodes = {
        machine_name: {"obj_type": "root", "is_schedulable": 0},
        "A": {"obj_type": "block", "is_schedulable": 1},
        "B": {"obj_type": "block", "is_schedulable": 1},
        "C": {"obj_type": "block", "is_schedulable": 1},
        "D": {"obj_type": "block", "is_schedulable": 1},
        "S0": {"obj_type": "switch", "is_schedulable": 1},
        "S1": {"obj_type": "switch", "is_schedulable": 1},
        "S2": {"obj_type": "switch", "is_schedulable": 1},
        "S3": {"obj_type": "switch", "is_schedulable": 1},
        "A0": {"obj_type": "block", "is_schedulable": 1, "idx": 0},
        "A1": {"obj_type": "block", "is_schedulable": 1, "idx": 1},
        "B0": {"obj_type": "block", "is_schedulable": 1, "idx": 2},
        "B1": {"obj_type": "block", "is_schedulable": 1, "idx": 3},
        "C0": {"obj_type": "block", "is_schedulable": 1, "idx": 4},
        "C1": {"obj_type": "block", "is_schedulable": 1, "idx": 5},
        "D0": {"obj_type": "block", "is_schedulable": 1, "idx": 6},
        "D1": {"obj_type": "block", "is_schedulable": 1, "idx": 7},
    }
    graph_edges = {
        machine_name: ("A", "B", "C", "D"),
        "A": ("S0", "S1"),
        "B": ("S1", "S2"),
        "C": ("S2", "S3"),
        "D": ("S3", "S0"),
        "S0": ("A0", "A1"),
        "S1": ("B0", "B1"),
        "S2": ("C0", "C1"),
        "S3": ("D0", "D1"),
    }

    def generate_bittree(
        self,
        validate=False,
        bitmask_class: type[EnhancedBitmaskLike] = EnhancedBitmask,
        debug=False,
    ):
        """create the graph using the nodes and edges structures."""
        graph: nx.DiGraph = nx.DiGraph()
        for node, node_dct in self.graph_nodes.items():
            graph.add_node(node, **node_dct)
        for node, edges in self.graph_edges.items():
            for node_dest in edges:
                graph.add_edge(node, node_dest)
        # show_nxgraph(graph)
        graph = apply_bitmask_to_nxgraph(
            self.machine_name,
            self.bit_total,
            graph,
            is_bitmask_indexable,
            sort_neighbors=self.graph_sort_neighbors,
            bitmask_class=bitmask_class,
            debug=debug,
        )
        return graph

    def get_parents(self, machine_name: str, node_name: str) -> list:
        if node_name == machine_name:
            nodes = []
        else:
            nodes = list(self.graph.predecessors(node_name))
        return nodes

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        if machine_name == node_name:
            obj_type = Obj_Type.ROOT.value
        elif node_name.startswith("S"):
            obj_type = Obj_Type.SWITCH.value
        else:
            obj_type = Obj_Type.BLOCK.value
        return obj_type


class Location_to_Mask_Dragonfly(ABC_Location_to_Mask_Graph):
    machine_name = MACHINE_NAME
    obj_types = [Obj_Type.ROOT.value, Obj_Type.SWITCH.value, Obj_Type.NODE.value]
    bit_total = 9
    graph_only = True
    graph_nodes = {
        machine_name: {"obj_type": "root", "is_schedulable": 0},
        "XSWglobal": {"obj_type": "switch", "is_schedulable": 1},
        "ASWlocal": {"obj_type": "switch", "is_schedulable": 1},
        "Ax": {"obj_type": "node", "is_schedulable": 1},
        "Ay": {"obj_type": "node", "is_schedulable": 1},
        "Az": {"obj_type": "node", "is_schedulable": 1},
        "BSWlocal": {"obj_type": "switch", "is_schedulable": 1},
        "Bx": {"obj_type": "node", "is_schedulable": 1},
        "By": {"obj_type": "node", "is_schedulable": 1},
        "Bz": {"obj_type": "node", "is_schedulable": 1},
        "CSWlocal": {"obj_type": "switch", "is_schedulable": 1},
        "Cx": {"obj_type": "node", "is_schedulable": 1},
        "Cy": {"obj_type": "node", "is_schedulable": 1},
        "Cz": {"obj_type": "node", "is_schedulable": 1},
    }
    graph_edges = {
        machine_name: ("XSWglobal",),
        "XSWglobal": (
            "ASWlocal",
            "BSWlocal",
            "CSWlocal",
        ),
        "ASWlocal": ("Ax", "Ay", "Az"),
        "BSWlocal": ("Bx", "By", "Bz"),
        "CSWlocal": ("Cx", "Cy", "Cz"),
    }

    def generate_bittree(
        self,
        validate=False,
        bitmask_class: type[EnhancedBitmaskLike] = EnhancedBitmask,
        debug=False,
    ):
        """create the graph using the nodes and edges structures."""
        graph: nx.DiGraph = nx.DiGraph()
        for node, node_dct in self.graph_nodes.items():
            graph.add_node(node, **node_dct)
        for node, edges in self.graph_edges.items():
            for node_dest in edges:
                graph.add_edge(node, node_dest)
        # show_nxgraph(graph)
        graph = apply_bitmask_to_nxgraph(
            self.machine_name,
            self.bit_total,
            graph,
            is_bitmask_indexable,
            sort_neighbors=self.graph_sort_neighbors,
            bitmask_class=bitmask_class,
            debug=debug,
        )
        return graph

    def get_parents(self, machine_name: str, node_name: str) -> list:
        if node_name == machine_name:
            nodes = []
        else:
            nodes = list(self.graph.predecessors(node_name))
        return nodes

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        if machine_name == node_name:
            obj_type = Obj_Type.ROOT.value
        elif "SW" in node_name:
            obj_type = Obj_Type.SWITCH.value
        else:
            obj_type = Obj_Type.NODE.value
        return obj_type


class Location_to_Mask_Dragonfly_Links(ABC_Location_to_Mask_Graph):
    machine_name = MACHINE_NAME
    obj_types = [
        Obj_Type.ROOT.value,
        Obj_Type.SWITCH.value,
        Obj_Type.LINK.value,
        Obj_Type.NODE.value,
    ]
    bit_total = 9
    graph_only = True
    graph_nodes = {
        machine_name: {"obj_type": "root", "is_schedulable": 0},
        "GL": {"obj_type": "link", "is_schedulable": 1},
        "S00": {"obj_type": "switch", "is_schedulable": 1},
        "Ax": {"obj_type": "node", "is_schedulable": 1},
        "Ay": {"obj_type": "node", "is_schedulable": 1},
        "Az": {"obj_type": "node", "is_schedulable": 1},
        "S01": {"obj_type": "switch", "is_schedulable": 1},
        "Bx": {"obj_type": "node", "is_schedulable": 1},
        "By": {"obj_type": "node", "is_schedulable": 1},
        "Bz": {"obj_type": "node", "is_schedulable": 1},
        "S02": {"obj_type": "switch", "is_schedulable": 1},
        "Cx": {"obj_type": "node", "is_schedulable": 1},
        "Cy": {"obj_type": "node", "is_schedulable": 1},
        "Cz": {"obj_type": "node", "is_schedulable": 1},
        "Cx<->Cy": {"obj_type": "link", "is_schedulable": 1},
        "Cy<->Cz": {"obj_type": "link", "is_schedulable": 1},
        "Cz<->Cx": {"obj_type": "link", "is_schedulable": 1},
    }
    graph_edges = {
        machine_name: ("GL",),
        "GL": (
            "S00",
            "S01",
            "S02",
        ),
        "S00": ("Ax", "Ay", "Az"),
        "S01": ("Bx", "By", "Bz"),
        "S02": ("Cx<->Cy", "Cy<->Cz", "Cz<->Cx", "Cx", "Cy", "Cz"),
        "Cx<->Cy": ("Cx", "Cy"),
        "Cy<->Cz": ("Cy", "Cz"),
        "Cz<->Cx": ("Cz", "Cx"),
    }

    def generate_bittree(
        self,
        validate=False,
        bitmask_class: type[EnhancedBitmaskLike] = EnhancedBitmask,
        debug=False,
    ):
        """create the graph using the nodes and edges structures."""
        graph: nx.DiGraph = nx.DiGraph()
        for node, node_dct in self.graph_nodes.items():
            graph.add_node(node, **node_dct)
        for node, edges in self.graph_edges.items():
            for node_dest in edges:
                graph.add_edge(node, node_dest)
        # show_nxgraph(graph)
        graph = apply_bitmask_to_nxgraph(
            self.machine_name,
            self.bit_total,
            graph,
            is_bitmask_indexable,
            sort_neighbors=self.graph_sort_neighbors,
            bitmask_class=bitmask_class,
            debug=debug,
        )
        return graph

    def get_parents(self, machine_name: str, node_name: str) -> list:
        if node_name == machine_name:
            nodes = []
        else:
            nodes = list(self.graph.predecessors(node_name))
        return nodes

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        if machine_name == node_name:
            obj_type = Obj_Type.ROOT.value
        elif node_name.startswith("S"):
            obj_type = Obj_Type.SWITCH.value
        elif "<->" in node_name or "GL" in node_name:
            obj_type = Obj_Type.LINK.value
        else:
            obj_type = Obj_Type.NODE.value
        return obj_type


def simple_generate_bittree(
    name,
    gl_sw_name,
    ll_delim,
    gll_delim,
    lsw_count,
    lno_count,
    sw_name_fmt,
    node_name_fmt,
):
    graph: nx.DiGraph = nx.DiGraph()
    graph.add_node(name, **{"obj_type": "root", "is_schedulable": 0})
    graph.add_node(gl_sw_name, **{"obj_type": "switch", "is_schedulable": 1})
    graph.add_edge(name, gl_sw_name)
    nodes = []
    node_edges = []
    switches = []
    for lsw_idx in range(lsw_count):
        sw_name = sw_name_fmt.format(lsw_idx=lsw_idx)
        switches.append(sw_name)
        graph.add_node(sw_name, **{"obj_type": "switch", "is_schedulable": 0})
        graph.add_edge(gl_sw_name, sw_name)
        node_lst = []
        for n_idx in range(lno_count):
            node_name = node_name_fmt.format(sw_name=sw_name, n_idx=n_idx)
            node_lst.append(node_name)
            nodes.append((node_name, {"obj_type": "node", "is_schedulable": 1}))
            node_edges.append((sw_name, node_name))
        # permutations are bidirectional, we only need one way
        ll_lst = [ll_delim.join(sorted(ll)) for ll in permutations(node_lst, 2)]
        ll_lst = sorted(list(set(ll_lst)))
        for link_name in ll_lst:
            graph.add_node(link_name, **{"obj_type": "link", "is_schedulable": 1})
            left, right = link_name.split(ll_delim, 1)
            graph.add_edge(sw_name, link_name)
            graph.add_edge(link_name, left)
            graph.add_edge(link_name, right)
    gll_lst = [gll_delim.join(sorted(gll)) for gll in permutations(switches, 2)]
    gll_lst = sorted(list(set(gll_lst)))
    for glink_name in gll_lst:
        graph.add_node(glink_name, **{"obj_type": "link", "is_schedulable": 1})
        left, right = glink_name.split(gll_delim, 1)
        graph.add_edge(gl_sw_name, glink_name)
        graph.add_edge(glink_name, left)
        graph.add_edge(glink_name, right)
    for node, node_dct in nodes:
        graph.add_node(node, **node_dct)
    for sw_name, node_name in node_edges:
        graph.add_edge(sw_name, node_name)
    return graph


class Location_to_Mask_Dragonfly_Full(ABC_Location_to_Mask_Graph):
    name = MACHINE_NAME
    machine_name = name
    obj_types = [
        Obj_Type.ROOT.value,
        Obj_Type.SWITCH.value,
        Obj_Type.LINK.value,
        Obj_Type.NODE.value,
    ]
    bit_total = 12
    graph_only = True

    def generate_bittree(
        self,
        validate=False,
        bitmask_class: type[EnhancedBitmaskLike] = EnhancedBitmask,
        debug=False,
    ):
        """create the graph using the nodes and edges structures."""
        name = self.name
        gl_sw_name = "GL"
        ll_delim = "<->"
        gll_delim = "<->"
        lsw_count = 3
        lno_count = 4
        sw_name_fmt = "S{lsw_idx:0>2}"
        node_name_fmt = "{sw_name}-N{n_idx:0>2}"
        graph = simple_generate_bittree(
            name,
            gl_sw_name,
            ll_delim,
            gll_delim,
            lsw_count,
            lno_count,
            sw_name_fmt,
            node_name_fmt,
        )
        if debug:
            for node in graph.nodes:
                print(node, graph.nodes[node])
        graph = apply_bitmask_to_nxgraph(
            self.machine_name,
            self.bit_total,
            graph,
            is_bitmask_indexable,
            sort_neighbors=self.graph_sort_neighbors,
            tail_retry_count=1,
            bitmask_class=bitmask_class,
            debug=debug,
        )
        return graph


class Location_to_Mask_LinkedAccelerators(ABC_Location_to_Mask_Graph):
    name = MACHINE_NAME
    machine_name = name
    obj_types = [
        Obj_Type.ROOT.value,
        Obj_Type.SWITCH.value,
        Obj_Type.LINK.value,
        Obj_Type.NODE.value,
    ]
    bit_total = 8
    graph_only = True

    def generate_bittree(
        self,
        validate=False,
        bitmask_class: type[EnhancedBitmaskLike] = EnhancedBitmask,
        debug=False,
    ):
        """create the graph using the nodes and edges structures."""
        name = self.name
        gl_sw_name = "GL"
        ll_delim = "<-pcie->"
        gll_delim = "<->"
        lsw_count = 2
        lno_count = 4
        sw_name_fmt = "S{lsw_idx:0>2}"
        node_name_fmt = "{sw_name}-N{n_idx:0>2}"
        graph = simple_generate_bittree(
            name,
            gl_sw_name,
            ll_delim,
            gll_delim,
            lsw_count,
            lno_count,
            sw_name_fmt,
            node_name_fmt,
        )
        if debug:
            for node in graph.nodes:
                print(node, graph.nodes[node])
        graph = apply_bitmask_to_nxgraph(
            self.machine_name,
            self.bit_total,
            graph,
            is_bitmask_indexable,
            sort_neighbors=self.graph_sort_neighbors,
            tail_retry_count=1,
            bitmask_class=bitmask_class,
            debug=debug,
        )
        return graph


class Location_to_Mask_Interconnect_Huge(ABC_Location_to_Mask_Graph):
    name = MACHINE_NAME
    machine_name = name
    obj_types = [
        Obj_Type.ROOT.value,
        Obj_Type.SWITCH.value,
        Obj_Type.LINK.value,
        Obj_Type.NODE.value,
    ]
    bit_total = 7391
    graph_only = True

    def generate_bittree(
        self,
        validate=False,
        bitmask_class: type[EnhancedBitmaskLike] = EnhancedBitmask,
        debug=False,
    ):
        """create the graph using the nodes and edges structures."""
        name = self.name
        gl_sw_name = "GL"
        ll_delim = "<->"
        gll_delim = "<->"
        lsw_count = 389
        lno_count = 19
        sw_name_fmt = "S{lsw_idx:04x}"
        node_name_fmt = "{sw_name}-N{n_idx:04x}"
        graph = simple_generate_bittree(
            name,
            gl_sw_name,
            ll_delim,
            gll_delim,
            lsw_count,
            lno_count,
            sw_name_fmt,
            node_name_fmt,
        )
        if debug:
            for node in graph.nodes:
                print(node, graph.nodes[node])
        graph = apply_bitmask_to_nxgraph(
            self.machine_name,
            self.bit_total,
            graph,
            is_bitmask_indexable,
            sort_neighbors=self.graph_sort_neighbors,
            tail_retry_count=1,
            bitmask_class=bitmask_class,
            debug=debug,
        )
        return graph


ltm_lst = [
    {
        "ts": pddt("1970-01-01"),
        "te": pddt("2018-01-01"),
        "ltm_class": XLocation_to_Mask_None,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2018-01-01"),
        "te": pddt("2019-01-01"),
        "ltm_class": Location_to_Mask_Simple,
        "pk": 15861864398923716641,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2019-01-01"),
        "te": pddt("2020-01-01"),
        "ltm_class": Location_to_Mask_Binary,
        "pk": 12955335836465270799,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2020-01-01"),
        "te": pddt("2021-01-01"),
        "ltm_class": Location_to_Mask_BinaryArray,
        "pk": 3959448401454836766,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2021-01-01"),
        "te": pddt("2021-02-01"),
        "ltm_class": Location_to_Mask_BinaryGraph,
        "pk": 2477230453084996305,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2021-02-01"),
        "te": pddt("2021-03-01"),
        "ltm_class": Location_to_Mask_Dragonfly,
        "pk": 7263386542337713672,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2021-03-01"),
        "te": pddt("2021-04-01"),
        "ltm_class": Location_to_Mask_Dragonfly_Links,
        "pk": 1168691695873863055,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2021-04-01"),
        "te": pddt("2021-05-01"),
        "ltm_class": Location_to_Mask_Dragonfly_Full,
        "pk": 12779943391514123908,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2021-05-01"),
        "te": pddt("2021-06-01"),
        "ltm_class": Location_to_Mask_LinkedAccelerators,
        "pk": 17181187861703342266,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2021-06-01"),
        "te": pddt("2022-01-01"),
        "ltm_class": Location_to_Mask_Interconnect_Huge,
        "pk": 63456533150571592,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2022-01-01"),
        "te": pddt("2038-01-01"),
        "ltm_class": XLocation_to_Mask_None,
        "name": MACHINE_NAME,
    },
]
