# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

# cython: language_level=3, boundscheck=True
import os
from typing import List

import networkx as nx

COMPARABLE_NODE_ATTRIBS = ["obj_type"]


def node_difference(
    graph_a: nx.DiGraph,
    node_a: str,
    graph_b: nx.DiGraph,
    node_b: str,
    compare_edges=False,
) -> List[str]:
    """compare two nodes

    compare parent
    compare children
    compare attributes of node_a vs node_b

    """
    node_a_children = graph_a.out_edges(node_a)
    node_b_children = graph_b.out_edges(node_b)
    node_a_parents = graph_a.in_edges(node_a)
    node_b_parents = graph_b.in_edges(node_b)
    node_a_dct = graph_a.nodes[node_a]
    node_b_dct = graph_b.nodes[node_b]
    node_a_bitmask = node_a_dct["bitmask"]
    node_b_bitmask = node_b_dct["bitmask"]
    differences = []
    len_children_a = len(node_a_children)
    len_children_b = len(node_b_children)
    for attrib in COMPARABLE_NODE_ATTRIBS:
        value_a = node_a_dct[attrib]
        value_b = node_b_dct[attrib]
        if value_a != value_a:
            differences.append(f"self:{node_a} difference of attr {attrib}: node_a({value_a}) != node_b({value_b})")
    if node_a_bitmask.length != node_b_bitmask.length:
        differences.append(
            f"self:{node_a} difference of bitmask size: self({node_a_dct['bitmask']}) != other({node_b_dct['bitmask']})"
        )
    else:
        if not (node_a_bitmask == node_b_bitmask).all():
            differences.append(
                f"self:{node_a} difference of bitmask: self({node_a_dct['bitmask']}) != other({node_b_dct['bitmask']})"
            )
    if compare_edges:
        for node_a_edge in node_a_children:
            if node_a_edge not in node_b_children:
                differences.append(f"child edge:{node_a_edge} is missing from node_b")
        for node_b_edge in node_b_children:
            if node_b_edge not in node_a_children:
                differences.append(f"child edge:{node_b_edge} is missing from node_a")
        for node_a_edge in node_a_parents:
            if node_a_edge not in node_b_parents:
                differences.append(f"parent edge:{node_a_edge} is missing from node_b")
        for node_b_edge in node_b_parents:
            if node_b_edge not in node_a_parents:
                differences.append(f"parent_edge:{node_b_edge} is missing from node_a")
        if len_children_a != len_children_b:
            differences.append(f"outdegree of node_a({len_children_a}) != node_b({len_children_b})")
    return differences


def validate_graph(graph: nx.DiGraph):
    directed = nx.is_directed(graph)
    dag = nx.is_directed_acyclic_graph(graph)
    if not directed or not dag:
        raise Exception(f"Graph is not valid: {directed=} {dag=}")


def show_nxgraph(graph, node_dct={}, edge_dct={}):
    from networkx.drawing.nx_agraph import graphviz_layout
    import matplotlib.pyplot as plt

    pos = graphviz_layout(graph, prog="dot")
    # pos = nx.spring_layout(graph)
    nx.draw_networkx_nodes(graph, pos, **node_dct)
    text = nx.draw_networkx_labels(graph, pos)
    for _, t in text.items():
        t.set_rotation(60)
    nx.draw_networkx_edges(graph, pos, edgelist=graph.edges, **edge_dct)
    plt.tight_layout()
    plt.show()


def write_nxgraph(graph, filename=None, node_dct={}):
    from networkx.drawing.nx_agraph import graphviz_layout
    import matplotlib.pyplot as plt

    pos = graphviz_layout(graph, prog="dot")
    # pos = nx.spring_layout(graph)
    nx.draw_networkx_nodes(graph, pos, **node_dct)
    nx.draw_networkx_labels(graph, pos)
    nx.draw_networkx_edges(graph, pos, edgelist=graph.edges, arrows=True)
    if not filename:
        filename = os.path.join(os.path.abspath("."), "graph.png")
    figure = plt.gcf()
    figure.set_size_inches(17, 11)
    print(f"writing out to {filename}")
    plt.savefig(filename, format="png", dpi=100)
