#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import os

import networkx as nx

from Octeres.util import superprint


def is_leaf(graph: nx.DiGraph, node):
    return graph.in_degree(node) >= 1 and graph.out_degree(node) == 0


def get_leaves(graph: nx.DiGraph, node):
    descendants = sorted(nx.algorithms.dag.descendants(graph, node))
    leaves = []
    for descendant in descendants:
        if is_leaf(graph, descendant):
            leaves.append(descendant)
    return leaves


class AnimatedGraph:
    """this is highly experimental and requires tweaking, but very useful for debugging."""

    def __init__(self, graph, layout, update_speed=200, save_gif_name=None):
        # FIXME: doesn't seem to animate everything when saving as a mp4
        import matplotlib.pyplot as plt

        self.plt = plt
        from matplotlib import animation

        self.graph: nx.DiGraph = graph
        self.layout = layout
        self.update_speed = update_speed  # ms
        self.fig, self.ax = self.plt.subplots(figsize=(24, 16))
        self.ani = animation.FuncAnimation(
            self.fig,
            self.update,
            repeat=False,
            interval=self.update_speed,
            fargs=(self.graph.size, layout, self.graph, self.ax),
        )
        self.title = ""
        self.save_gif_name = save_gif_name
        self.ani_path = []
        self.fps = 30
        self.add_count = 10

    def update(self, count, n, layout, graph, ax):
        if len(self.ani_path) <= 0:
            self.ani.event_source.stop()
        else:
            ax.clear()
            nx.draw_networkx_edges(graph, pos=layout, ax=ax, edge_color="gray")
            node, color = self.ani_path.pop(0)
            node_lst = set(graph.nodes()) - {node}
            null_nodes = nx.draw_networkx_nodes(graph, pos=layout, nodelist=node_lst, node_color="white", ax=ax)
            null_nodes.set_edgecolor("black")
            path = nx.draw_networkx_nodes(graph, pos=layout, nodelist={node}, node_color=color, ax=ax)
            path.set_edgecolor("blue")
            nx.draw_networkx_edges(graph, pos=layout, ax=ax, edge_color="gray")
            if self.title:
                ax.set_title(f"{self.title}")
            nx.draw_networkx_labels(graph, layout)

    def add(self, node, color="r"):
        """write out the data n times to actually see it."""
        for _ in range(self.add_count):
            self.ani_path.append(
                (
                    node,
                    color,
                )
            )

    def show(self):
        if self.save_gif_name:
            full_path = os.path.join(os.path.abspath("."), self.save_gif_name)
            superprint(f"writing to {full_path}")
            self.ani.save(full_path, writer="ffmpeg", fps=self.fps)
        else:
            self.plt.show()
        self.plt.close(self.fig)
