#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

"""
CLI for embedding the source code of modules imported by a Python script into a new instance of that script
"""

# pylint: disable=missing-class-docstring

from dataclasses import dataclass, field
import logging
from pathlib import Path
from typing import List, TextIO
import click

import Octeres.embed
from Octeres.embed.source import get_module_source
from Octeres.embed.generate import generate_embedded_script

logging_already_initialized = logging.getLogger().hasHandlers()
if logging_already_initialized:
    logging.basicConfig()
log = logging.getLogger(__name__)


@dataclass
class GlobalSettings:
    verbose: int = 0
    search_dirs: List[Path] = field(default_factory=list)


pass_global_settings = click.make_pass_decorator(GlobalSettings, ensure=True)


@click.group()
@pass_global_settings
@click.option(
    "--search-dir",
    "-d",
    help="directory to search for embeddable Python modules",
    type=click.Path(exists=True, file_okay=False, path_type=Path),
    multiple=True,
)
@click.option("--verbose", "-v", help="Increase verbosity level", count=True)
def cli(global_settings: GlobalSettings, verbose: int, search_dir: List[Path]) -> None:
    """
    Process global CLI options
    """
    global_settings.verbose = verbose
    global_settings.search_dirs = search_dir
    if not logging_already_initialized:
        # if logging was already initialized, then we are likely testing and should use existing logging level
        if global_settings.verbose >= 2:
            log.setLevel(logging.DEBUG)
        elif global_settings.verbose == 1:
            log.setLevel(logging.INFO)
        else:
            log.setLevel(logging.WARN)
        Octeres.embed.set_logging_level(log.level)


@cli.command()
@pass_global_settings
@click.argument(
    # TODO: Allow a script to be provided via stdin
    # TODOL verify that the script is valid python for the Python interpreter running this script
    "input-script",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option(
    "--output-script",
    "--out",
    "-o",
    help="Output generated script to FILENAME (default: stdout)",
    type=click.File("w"),
    required=False,
    default="-",
)
@click.option(
    "--restore-sys",
    help="Restore sys.modules and sys.meta_path when script exits (useful for multiple evals)",
    type=click.BOOL,
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--new-context",
    "--new-ctx",
    "-d",
    help="Duplicate the current context (globals/locals) before enabling the embedded importer",
    type=click.BOOL,
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--cache-modules",
    help="Cache embedded modules that have already been imported for use during another eval",
    type=click.BOOL,
    is_flag=True,
    default=False,
    show_default=True,
)
def generate(global_settings: GlobalSettings, input_script: Path, output_script: TextIO, **kwargs) -> None:
    """
    Create script with embedded modules
    """
    in_source = get_module_source(input_script)
    out_module = generate_embedded_script(in_source, global_settings.search_dirs, **kwargs)
    log.info("Writing script with embedded modules to %s", output_script.name)
    output_script.write("\n".join(out_module.text))


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    cli()
